from tempfile import NamedTemporaryFile

import pandas as pd
import numpy as np
from tensorflow.keras import backend as K
from tensorflow.keras.layers import Input, Flatten, Dense, Dropout, Lambda, concatenate,Add, Multiply, RepeatVector, Reshape
from tensorflow.keras.models import Model
from tensorflow.keras.models import load_model
from tensorflow.keras.wrappers.scikit_learn import KerasRegressor
from numpy.random import seed
from sklearn.metrics import make_scorer
from sklearn.model_selection import GridSearchCV
from tensorflow.keras.losses import mean_squared_error
from tensorflow.keras.activations import softmax, exponential
from tensorflow.keras.layers import Softmax
from sklearn.model_selection import KFold, ParameterGrid
from mr_uplift.erupt import get_weights, erupt
import tensorflow as tf

def reduce_concat(x):
    """concatenates object into one string
    Args:
      x (array): values to concatenate into one string
    Returns:
      string of objects
    """
    return np.array(['_'.join(map(str, q)) for q in x])

def mse_loss_multi_output_np(yTrue, yPred):
    """Returns Loss for Keras. This is essentially a multi-output regression that
    ignores missing data. It's similar in concept to matrix factorization with
    missing data in that it only includes non missing values in loss function.
    Anything with -999.0 is ignored
    Args:
        yTrue (keras array): True Values
        yPred (keras array): Predictions
    Returns:
        Function to minimize
    """
    output = (yTrue  - yPred )**2
    return np.mean(output)


mse_loss_multi_output_scorer = make_scorer(
    mse_loss_multi_output_np, greater_is_better=False)


def hete_mse_loss_multi_output(
        input_shape,
        output_shape,
        num_nodes=256,
        dropout=.5,
        num_layers=1,
        activation='relu'):
    """Returns standard neural network
    Args:
        input_shape (int): number of features in training dataset
        num_nodes (int): Number of nodes in layers
        dropout (real): dropout percentage
        num_layers (real): number of num_layer
        activation (keras function): nonlinear activation function
    Returns:
        Keras Model
    """
    input = Input(shape=(input_shape,))

    x = Dense(num_nodes, activation=activation)(input)
    x = Dropout(dropout)(x)

    if(num_layers > 1):
        for q in range(num_layers - 1):
            x = Dense(num_nodes, activation=activation)(x)
            x = Dropout(dropout)(x)

    x = Dense(output_shape)(x)

    model = Model(input, x)
    model.compile(
        optimizer='rmsprop',
        loss=mean_squared_error,
        metrics=[mean_squared_error])

    return model


def train_model_multi_output_w_tmt(x, y, param_grid=None, n_jobs=-1, cv=5):
    """Trains and does gridsearch on a keras nnet. Response can be multi output
    Args:
        x (np array): explanatory variables
        y (np array): response variables. could be more than 1
        n_jobs (int): number of cores to run on.
        cv (int): number of cross vaildation_folds
    Returns:
        Gridsearched multi output model
    """

    if param_grid is None:
        param_grid = dict(num_nodes=[8, 128], dropout=[.1, .25, .5], activation=[
                          'relu'], num_layers=[1, 2], epochs=[50], batch_size=[400, 4000, 8000])

    model = KerasRegressor(
        build_fn=hete_mse_loss_multi_output,
        input_shape=x.shape[1],
        output_shape=y.shape[1],
        num_nodes=32,
        num_layers=1,
        epochs=100,
        verbose=False,
        dropout=.2,
        batch_size=1000)

    grid = GridSearchCV(
        estimator=model,
        param_grid=param_grid,
        n_jobs=n_jobs,
        scoring=mse_loss_multi_output_scorer,
        verbose=False,
        cv=cv)

    grid_result = grid.fit(x, y)

    return grid_result


def create_base_network(input_shape, num_responses,
  num_layers = 1, num_nodes = 8, dropout = .1, activation = 'relu'):
  '''Base network to be shared (eq. to feature extraction).
  '''
  input = Input(shape=(input_shape,))

  x = Dense(num_nodes, activation=activation)(input)
  x = Dropout(dropout)(x)

  if(num_layers > 1):
     for q in range(num_layers - 1):
         x = Dense(num_nodes, activation=activation)(x)
         x = Dropout(dropout)(x)

  x = Dense(num_responses)(x)

  model = Model(input, x, name = 'net_model')

  return model


def optim_loss(yTrue,yPred):
    """Returns Loss for Keras. This function essentially asks the question:
    'what is the probability that a particular treatment yields the maximum?'.
    It does this by multiplying a softmax probability to the response variable
    of interest. If a response is missing then it is not included in loss.
    Args:
        yTrue (keras array): True Values
        yPred (keras array): Predictions
    Returns:
        Function to minimize
    """

    output = yTrue*yPred
    output = K.sum(output,axis = -1)
    return -K.mean(output)


def missing_mse_loss_multi_output(yTrue,yPred):
    """Returns Loss for Keras. This is essentially a multi-output regression that
    ignores missing data. It's similar in concept to matrix factorization with
    missing data in that it only includes non missing values in loss function.
    Args:
        yTrue (keras array): True Values
        yPred (keras array): Predictions
    Returns:
        Function to minimize
    """
    equals = K.cast(K.not_equal(yTrue, K.cast_to_floatx(-999.0)), dtype='float32')
    output = (yTrue * equals  - yPred*equals )**2
    return K.mean(output)


def create_mo_optim_model(input_shape, num_responses, unique_treatments, num_layers,
  num_nodes, dropout, activation, alpha):
    """Returns model multiple-response uplift optimization model.
    Args:
        input_shape (int): number of features in training dataset
        num_responses (int): number of response variables
        unique_treatments (np.array): array of unique treatments
        num_layers (real): number of num_layer
        num_nodes (int): Number of nodes in layers
        dropout (real): dropout percentage
        activation (keras function): nonlinear activation function
        alpha (float): number between 0 and 1. 1 corresponds to full optimization
        0 corresponds to missing MSE loss and should be similiar to `train_model_multi_output_w_tmt`
        function output. intended to be gridsearched over
    Returns:
        Keras Model
    """

    inputs_x = Input(shape=(input_shape,))
    util_weights = Input(shape=(num_responses,))
    masks = Input(shape=(unique_treatments.shape[0],))

    batch_size = K.shape(inputs_x)[0]

    unique_tmts = [K.variable(x.reshape(1,unique_treatments.shape[1])) for x in unique_treatments]
    unique_tmts_full = [K.tile(x, (batch_size,1)) for x in unique_tmts]

    x_ts = [concatenate([tmts_full, inputs_x]) for tmts_full in unique_tmts_full]

    model = create_base_network(input_shape+unique_treatments.shape[1] ,num_responses,
    num_layers, num_nodes, dropout, activation )

    outputs = [model(x) for x in x_ts]

    outputs_mse = [RepeatVector(1)(data) for data in outputs]
    outputs_mse = concatenate(outputs_mse, axis = 1, name = 'outputs_mse')

    outputs_weighted = [Multiply()([outs,util_weights]) for outs in outputs]
    #expand to 3 dimensions. Used to concatenate in next line.
    outputs_weighted = [RepeatVector(1)(data) for data in outputs_weighted]
    outputs_weighted = concatenate(outputs_weighted, axis = 1)

    util_by_tmts = Lambda(lambda x: K.sum(x, axis=-1))(outputs_weighted)

    masked_inp = tf.where(masks!=0, util_by_tmts, tf.float64.min)
    util_by_tmts_prob = Lambda(lambda x: softmax(x), name = 'util_by_tmts_prob')(util_by_tmts)

    model = Model([inputs_x, util_weights, masks], [util_by_tmts_prob, outputs_mse])
    model.compile(optimizer='rmsprop',
                loss={'util_by_tmts_prob':optim_loss, 'outputs_mse': missing_mse_loss_multi_output},
               loss_weights={'util_by_tmts_prob': alpha, 'outputs_mse': (1-alpha)})

    return model


def copy_data(x,y,t, masks, weights, n):
    """Copys input data n times. Found to be helpul in building create_mo_optim_model
    Args:
        x (arrary): input data
        y (arrary): response data
        t (arrary): treatment data
        n (int): number of times to copy
    Returns:
        copied data
    """
    x = np.concatenate([x.copy() for q in range(n)])
    y = np.concatenate([y.copy() for q in range(n)])
    t = np.concatenate([t.copy() for q in range(n)])
    masks = np.concatenate([masks.copy() for q in range(n)])
    weights = np.concatenate([weights.copy() for q in range(n)])

    return x, y, t, masks, weights


def get_random_weights(y, random_seed = 22):
    """Creates a random uniform weight matrix
    Args:
        y (arrary): response data
    Returns:
        random weights of dimentions y
    """
    np.random.seed(random_seed)
    n_obs = y.shape[0]
    n_responses = y.shape[1]
    if n_responses > 1:
        utility_weights = np.random.uniform(0,1,size = n_obs*n_responses).reshape(n_obs,n_responses)
        utility_weights = utility_weights/np.abs(utility_weights).sum(axis=1).reshape(-1,1)
        utility_weights_neg = 1-2*(np.concatenate([np.random.uniform(0,1,n_obs).reshape(-1,1) for q in range(n_responses)], axis = 1)<.5)
        utility_weights = utility_weights*utility_weights_neg
    else:
        utility_weights = np.ones(n_obs).reshape(-1,1)
    return utility_weights

def treatments_to_text(t, unique_treatments):
    if(unique_treatments.shape[1] > 1):
        str_t = reduce_concat(t)
        str_unique_treatments = reduce_concat(unique_treatments)
    else:
        str_t = [str(q) for q in np.squeeze(t)]
        str_unique_treatments = ([str(q) for q in np.squeeze(unique_treatments)])

    return str_t, str_unique_treatments


def prepare_data_optimized_loss(x, y , t, mask, unique_treatments, weights = None,
copy_several_times = None, random_seed = 22):
    """Prepares dataset to be used in `create_mo_optim_model` model build.
    Args:
        x (np array): explanatory variables
        y (np array): response variables
        t (np array): treatment variable
        unique_tmts (np array): unique treatment variables
        weight observations inverse proportional to frequency of treatment
        copy_several_times (int): number of times to repeat dataset and create random weights
        random_seed (int): seed for rng
    Returns:
        x (np array): explanatory variables. Potentially repeated severa times.
        utility_weights (np array): random utility weights for each observation
        missing_utility (np array): new weighted sum response matrix. Has non-zero
        elements for treatment column they saw
        missing_y_mat (np array): 3-d array of num_obs, num_tmts, num_response variables.
        has non -999 values for
    """

    if weights is None:
        weights = np.ones(y.shape[0])

    if copy_several_times is not None:
        x,y,t, mask, weights = copy_data(x,y,t,mask, weights, copy_several_times)

    str_t, str_unique_treatments = treatments_to_text(t, unique_treatments)

    #used to maintain order of treatments
    treatments_order = dict(zip(str_unique_treatments,  range(len(str_unique_treatments))))

    #creates 3-d matrix num_observations, num_treatments, num_responses. Has response variables
    #for treatment locations that observation was assigned. Else it recievd -999. This is a special value for loss function
    missing_y_mat = np.zeros((y.shape[0], len(unique_treatments), y.shape[1])) - 999
    for key, value in treatments_order.items():
      temp_locs = np.where(key == np.array(str_t))[0]
      missing_y_mat[temp_locs, value,:] = np.array(y)[temp_locs,:]

    #creates new response variable
    utility_weights = get_random_weights(y, random_seed = random_seed)
    utility_y = (utility_weights*np.array(y)).sum(axis=1)
    #get weights of treatments if treatments not uniform

    utility_y = utility_y.reshape(-1,1)*weights.reshape(-1,1)

    #Creates an num_obs by num_treatments mat. Has new weighted utlity response
    #if user recieved treatment, else 0
    missing_utility = pd.get_dummies(pd.DataFrame(str_t).iloc[:,0])
    missing_utility = np.array(missing_utility[[k for k,v in treatments_order.items()]])
    missing_utility = utility_y.reshape(-1,1)*missing_utility

    mask = np.array(mask,dtype=bool)

    return x, utility_weights, missing_utility, missing_y_mat, mask, weights


def gridsearch_mo_optim(x, y, t, observation_weights, mask_tmt_locations, n_splits=5,  param_grid=None):
    """Gridsearches over create_mo_optim_model. Since it has multiple inputs/ outputs
    I 'rolled' my own.

    Args:
        x (np array): explanatory variables
        y (np array): response variables
        t (np array): treatment variable
        param_grid (dictionary): parameter grid values
        use_propensity (Boolean): If True will build propensity model and weight observations accordingly
        propensity_cutoff_weight (float): Anything greater than this propensity weight will be masked
    Returns:
        mod (keras model): Fitted model with best hyperparameters
        optim_grid (dictionary): best hyperparameters in dictionary
        model_score (float): best score of data
        propensity_model (rf classifier): Propensity model only if use_propensity is True
    """

    unique_treatments = np.unique(t, axis = 0)

    if len(unique_treatments.shape) == 1:
        unique_treatments = unique_treatments.reshape(unique_treatments.shape[0], -1)

    grid = ParameterGrid(param_grid)

    kf = KFold(n_splits=n_splits, shuffle=True, random_state=22)
    kf.get_n_splits(y)

    results = []
    for params in grid:
        copy_several_times = params['copy_several_times']
        temp_results = []


        for train_index, test_index in kf.split(y):

            weights_train = observation_weights[train_index]
            weights_test = observation_weights[test_index]

            mask_train = mask_tmt_locations[train_index]
            mask_test = mask_tmt_locations[test_index]


            x_train, utility_weights_train, new_response_train, big_y_train, mask_train, weights_train  = prepare_data_optimized_loss(x[train_index],y[train_index],
            t[train_index], mask = mask_train, unique_treatments = unique_treatments,  weights = weights_train, copy_several_times = copy_several_times)
            x_test, utility_weights_test, new_response_test, big_y_test, mask_test, weights_test   = prepare_data_optimized_loss(x[test_index], y[test_index],
            t[test_index], mask = mask_test, unique_treatments = unique_treatments, weights = None, copy_several_times = None)



            mod = create_mo_optim_model(input_shape = x.shape[1],
              num_responses = y.shape[1],
              unique_treatments = unique_treatments,
              num_layers = params['num_layers'],
              num_nodes = params['num_nodes'],
              dropout = params['dropout'],
              alpha = params['alpha'],
              activation = params['activation'])

            mod.fit([x_train, utility_weights_train, mask_train], [new_response_train, big_y_train],
            epochs = params['epochs'],
            batch_size = params['batch_size'],
            verbose = False)

            preds = mod.predict([x_test, utility_weights_test,mask_test])[0]

            preds_masked = preds*mask_test - (1-mask_test)*10**6
            optim_value_location = np.argmax(preds_masked, axis = 1)
            tmt_location = np.argmax(np.abs(new_response_test), axis = 1)

            erupts = erupt(np.array(new_response_test).sum(axis=1).reshape(-1,1), tmt_location, optim_value_location, weights=weights_test, names=None)

            temp_results.append(erupts['mean'])

        results.append(np.mean(temp_results).mean())

    optim_grid = [x for x in grid][np.argmax(np.array(results))]

    mod  = create_mo_optim_model(input_shape = x.shape[1],
            num_responses = y.shape[1],
            unique_treatments = unique_treatments, num_layers = optim_grid['num_layers'],
            num_nodes = optim_grid['num_nodes'],
            dropout = optim_grid['dropout'],
            alpha = optim_grid['alpha'],
            activation = optim_grid['activation'])

    x, utility_weights, new_response, big_y, mask_train, weights_train  = prepare_data_optimized_loss(x,y,t,mask_tmt_locations, unique_treatments,
    weights = observation_weights, copy_several_times = optim_grid['copy_several_times'])


    mod.fit([x, utility_weights, mask_train] , [new_response, big_y], epochs = optim_grid['epochs'],
        batch_size = optim_grid['batch_size'], verbose = False)

    return mod, optim_grid, results[np.argmax(np.array(results))]



import numpy as np
import pandas as pd
import dill
import copy
from tensorflow.keras.models import load_model
from mr_uplift.keras_model_functionality import train_model_multi_output_w_tmt, gridsearch_mo_optim, get_random_weights, treatments_to_text
from mr_uplift.erupt import get_erupts_curves_aupc, get_best_tmts, erupt, get_weights, softmax

from keras.wrappers.scikit_learn import KerasRegressor
from sklearn.model_selection import GridSearchCV
from sklearn.preprocessing import StandardScaler, KBinsDiscretizer, FunctionTransformer
from sklearn.decomposition import PCA
from sklearn.model_selection import train_test_split
from mr_uplift.calibrate_uplift import UpliftCalibration
from sklearn.pipeline import make_pipeline
from sklearn.ensemble import RandomForestClassifier

def get_t_data(values, num_obs):
    """Repeats treatment to several rows and reshapes it. Used to append treatments
    to explanatory variable dataframe to predict counterfactuals.
    Args:
      values (array): treatments to predict
    Returns:
      repeated treatment values
    """

    if isinstance(values, np.ndarray):
        len_values = len(values)
    else:
        len_values = 1

    values = np.array(values)
    values_rep = np.full((num_obs, len_values), values.reshape(1, len_values))

    return values_rep


def reduce_concat(x):
    """concatenates object into one string
    Args:
      x (array): values to concatenate into one string
    Returns:
      string of objects
    """
    return np.array(['_'.join(map(str, q)) for q in x])


class MRUplift(object):
    def __init__(self, **kw):

        """I'm putting this in here because I think its necessary for .copy()
        function later. Not sure if thats the case.
        """
        self.kw = kw
        self.__dict__.update(kw)

    def fit(self, x, y, t, test_size=0.7, random_state=22, param_grid=None,
            n_jobs=-1, cv=5, optimized_loss = False, PCA_x = False, PCA_y = False, bin = False,
            use_propensity = False, propensity_score_cutoff = 100):
        """Fits a Neural Network Model of the form y ~ f(t,x). Creates seperate
        transformers for y, t, and x and scales each. Assigns train / test split.

        Args:
        x (np array or pd.dataframe): Explanatory Data of shape num_observations
          by num_explanatory_variables
        y (np array or pd.dataframe): Response Variables of shape
          num_observations by num_response_variables
        t (np array or pd.dataframe): Treatment Variables of shape
          num_observations by num_treatment columns
        test_size (float): Percentage of observations to be in test set
        random_state (int): Random state for train test split. This is used in other parts
        of class.
        param_grid (dict): Parameters of keras model to gridsearch over.
        n_jobs (int): number of cores to run on.
        cv (int): number of cross vaildation_folds
        optimized_loss (boolean): If True will use the optimized loss funcionality.
        PCA_x (boolean): If True it will use PCA to preprocess explanatory variables
        PCA_y (boolean): If True it will use PCA to preprocess response variables
        use_propensity (boolean): If True will use propensity scores from a RF. Best for observational data.
        propensity_score_cutoff (float): maximum weight of propensity score. If too high than it wouldn't have had
          much support in original model and should probably be exlcuded.

        Returns:
        Builds a neural network and assigns it to self.model
        """

        self.unique_t = np.unique(np.array(t).astype('float'), axis=0)

        self.num_t = len(np.unique(t))
        self.num_responses = y.shape[1]

        self.x = np.array(x).astype('float')
        self.y = np.array(y).astype('float')
        self.t = np.array(t).astype('float')

        self.test_size = test_size
        self.random_state = random_state
        self.propensity_score_cutoff = propensity_score_cutoff

        if isinstance(x, pd.DataFrame):
            self.x_names = x.columns.values
        else:
            self.x_names = None

        if isinstance(y, pd.DataFrame):
            self.y_names = y.columns.values
        else:
            self.y_names = None

        if isinstance(t, pd.DataFrame):
            self.t_names = t.columns.values
        else:
            self.t_names = None

        #if no holdout set the test_size to zero
        if self.test_size == 0:
            x_train = x
            y_train = y
            t_train = t
        else:
            x_train, x_test, y_train, y_test, t_train, t_test = train_test_split(
                self.x, self.y, self.t, test_size=self.test_size,
                random_state=self.random_state)

        #I've found that scaling features helps estimate nets.
        #Using PCA of features space should also help.

        if PCA_y:
            self.y_ss = PCA(whiten = True)
        else:
            self.y_ss = StandardScaler()


        if PCA_x:
            self.x_ss = PCA(whiten = True)

        elif bin:
            self.x_ss = make_pipeline(
                     KBinsDiscretizer(n_bins = 10),
                     FunctionTransformer(lambda x: x.todense(), accept_sparse=True)
                )
        else:
            self.x_ss = StandardScaler()

        self.t_ss = StandardScaler()

        self.y_ss.fit(y_train)
        self.x_ss.fit(x_train)
        self.t_ss.fit(t_train)

        y_train_scaled = self.y_ss.transform(y_train)
        x_train_scaled = self.x_ss.transform(x_train)
        t_train_scaled = self.t_ss.transform(t_train)

        x_t_train = np.concatenate([t_train_scaled, x_train_scaled], axis=1)


        str_t_train, str_unique_treatments = treatments_to_text(t_train.astype('float'), self.unique_t)

        if optimized_loss:
            if use_propensity:

                param_grid_propensity = {
                'n_estimators': [500],
                'max_features': ['auto'],
                'max_depth': [1,2,4,8],
                'oob_score' : [True],
                'n_jobs' : [-1]
                    }

                propensity_model = GridSearchCV(estimator=RandomForestClassifier(max_depth = 8, n_jobs = 1, oob_score = True, n_estimators = 500),
                    param_grid=param_grid_propensity, cv=3, scoring='neg_log_loss')
                propensity_model.fit(x_train_scaled, str_t_train)

                self.propensity_model = propensity_model.best_estimator_
                propensity_model = propensity_model.best_estimator_

                propensity_scores = pd.DataFrame(1/(propensity_model.oob_decision_function_+np.finfo(float).eps))
                propensity_scores.columns = propensity_model.classes_

                mask_tmt_locations = np.array((propensity_scores < self.propensity_score_cutoff)*1)

                str_t_series = pd.Series(str_t_train)
                observation_weights = np.array(propensity_scores.lookup(str_t_series.index, str_t_series.values)).reshape(-1,1)

            else:
                self.propensity_model = None
                mask_tmt_locations = np.ones(t_train.shape[0]*len(self.unique_t)).reshape(t_train.shape[0], len(self.unique_t))
                observation_weights = get_weights(str_t_train)

            keep_locs_on_observations = np.where(observation_weights > self.propensity_score_cutoff/(self.propensity_score_cutoff-1))[0]

            net = gridsearch_mo_optim(x_train_scaled[keep_locs_on_observations],
                                      y_train_scaled[keep_locs_on_observations],
                                      t_train_scaled[keep_locs_on_observations],
                         param_grid=param_grid,
                         n_splits=cv,
                         observation_weights=observation_weights[keep_locs_on_observations],
                         mask_tmt_locations=mask_tmt_locations[keep_locs_on_observations])

            self.best_score_net = net[2]
            self.best_params_net = net[1]
            #only need embedded layer and not whole net
            self.model = net[0].get_layer('net_model')

        else:
            net = train_model_multi_output_w_tmt(x_t_train, y_train_scaled,
                                                 param_grid=param_grid,
                                                 n_jobs=n_jobs,
                                                 cv=cv)

            self.best_score_net = net.best_score_
            self.best_params_net = net.best_params_
            self.model = net.best_estimator_.model
            self.propensity_model = None


    def predict(self, x, t, response_transformer = False):
        """Returns predictions of the fitted model. Transforms both x and t then
        concatenates those two variables into an array to predict on. Finally,
        an inverse transformer is applied on predictions to transform to original
        response means and standard deviations.

        Args:
        x (np array or pd.dataframe): Explanatory Data of shape num_observations
          by num_explanatory_variables
        t (np array or pd.dataframe): Treatment Variables of shape
          num_observations by num_treatment columns
        response_transformer (boolean): If true will use the trained scaler to transform
          responses. I've noticed that using this in production degrades performance
          becuase model optimizes scaled data.

        Returns:
        Predictions fitted model
        """

        x_t_new = np.concatenate(
            [self.t_ss.transform(t), self.x_ss.transform(x)], axis=1)
        preds = self.model.predict(x_t_new)

        if response_transformer:
            preds = self.y_ss.inverse_transform(preds)

        return preds

    def predict_ice(self, x=None, treatments=None, calibrator=False,
        response_transformer = False):
        """Predicts all counterfactuals with new data. If no new data is
            assigned it will use test set data. Can subset to particular treatments
            using treatment assignment. Can also apply calibrator function (experimental)
            to predictions.

            The 'ice' refers to Individual Conditional Expectations. A better
            description of this can be found here:
            https://arxiv.org/pdf/1309.6392.pdf
        Args:
          x (np.array): new data to predict. Will use test data if not given
          treatments (np.array): Treatments to predict on. If none assigned then
          original training treatments are used.
        calibrator (boolean): If true will use the trained calibrator to transform
          responses. Otherwise will use the response inverse transformer
        response_transformer (boolean): If true will use the trained scaler to transform
          responses. I've noticed that using this in production degrades performance
          becuase model optimizes scaled data.
        Returns:
          Counterfactuals for all treatments and response variables. an arrary of
          num_treatments by num_observations by num_responses
        """

        if x is None:
            x_train, x_test, y_train, y_test, t_train, t_test = train_test_split(
                self.x, self.y, self.t, test_size=self.test_size,
                random_state=self.random_state)
        else:
            x_test = x

        if treatments is None:
            treatments = self.unique_t

        ice = np.array([self.predict(x_test, get_t_data(
            t, x_test.shape[0]), response_transformer = response_transformer) for t in treatments])

        if calibrator:
            ice = self.calibrator.transform(ice)

        return ice

    def get_erupt_curves(self, x=None, y=None, t=None, objective_weights=None,
                         treatments=None, calibrator=False,
                         response_transformer = False,
                         propensity_score_cutoff = 100):
        """Returns ERUPT Curves and distributions of treatments. If either x or
        y or t is not inputted it will use test data.

        If there is only one response variable then it will assume we want to maximize the response.
        It will calculate and return the ERUPT metric and distribution of treatments.
        An introduction to ERUPT metric can be found here https://medium.com/building-ibotta/erupt-expected-response-under-proposed-treatments-ff7dd45c84b4

        If there are mutliple responses it will use objective_weights to create a weighted sum of response
        variables. It will then use this new weighted sum to determine optimal treatments and calculate ERUPT
        metrics accordingly. Repeatedly doing this with different weights will lead estimations of tradeoffs.

        ERUPT curves are described in more detail here:
        https://medium.com/building-ibotta/estimating-and-visualizing-business-tradeoffs-in-uplift-models-80ff845a5698

        Args:
          x (np.array): new data to predict. Will use test data if not given
          y (np.array): responses
          t (np.array): treatments
          objective_weights (np.array): of dim (num_weights by num_response_variables).
          if none is assigned it will trade off costs of first two response variables and
          assume that first column is to be maximized and second column is to be minimized
          treatments (np.array): treatments to use in erupt calculations
          calibrator (boolean): If true will use the trained calibrator to transform
            responses. Otherwise will use the response inverse transformer
          response_transformer (boolean): If true will use the trained scaler to transform
            responses. I've noticed that using this in production degrades performance
            becuase model optimizes scaled data.

        Returns:
          ERUPT Curves and Treatment Distributions
        """

        if self.num_responses == 1:
            objective_weights = np.array([1]).reshape(1,-1)

        if objective_weights is None:
            objective_weights = np.zeros((11, self.num_responses))
            objective_weights[:, 0] = np.linspace(0.0, 1.0, num=11)
            objective_weights[:, 1] = -np.linspace(1.0, 0.0, num=11)

        if any([q is None for q in [x, y, t] ]):
            print("Using Test Data Set")
            x_train, x, y_train, y, t_train, t = train_test_split(
                self.x, self.y, self.t, test_size=self.test_size,
                random_state=self.random_state)

        if treatments is None:
            treatments = self.unique_t

        x = np.array(x).astype('float')
        t = np.array(t).astype('float')
        y = np.array(y).astype('float')


        str_t, str_unique_treatments = treatments_to_text(t, treatments)

        to_keep_locs = np.where([z in str_unique_treatments for z in str_t])[0]

        y = y[to_keep_locs]
        t = t[to_keep_locs]
        x = x[to_keep_locs]
        str_t = np.array(str_t)[to_keep_locs]

        ice_preds = self.predict_ice(x, treatments, calibrator, response_transformer)


        if self.propensity_model is not None:

            prob_tmts = self.propensity_model.predict_proba(self.x_ss.transform(x))+np.finfo(float).eps
            propensity_scores = pd.DataFrame(1/prob_tmts)
            propensity_scores.columns = self.propensity_model.classes_
            propensity_scores = propensity_scores[str_unique_treatments]

            mask_tmt_locations = np.array((propensity_scores < self.propensity_score_cutoff)*1)

            t_series = pd.Series(str_t)
            observation_weights = propensity_scores.lookup(t_series.index, t_series.values)

        else:

            mask_tmt_locations = np.ones(t.shape[0]*len(treatments)).reshape(t.shape[0], len(treatments))
            observation_weights = get_weights(str_t)

        return get_erupts_curves_aupc(
            y,
            np.array(str_t),
            ice_preds,
            str_unique_treatments,
            objective_weights,
            names=self.y_names,
            mask_tmt_locations = mask_tmt_locations,
            observation_weights = np.array(observation_weights))

    def copy(self):
        """Copies MRUplift Class. Not sure if this is the best way but it
        works.
        """
        return copy.copy(self)

    def save(self, file_location):
        """Saves MRUplift Class to location. Will save two outputs:
        keras model and rest of MRUplift class.

        Args:
          file_location (str): File location to save data
        Returns:
          Nothin. Saves file to location
        """
        model = self.model
        uplift_class_copy = self.copy()
        uplift_class_copy.model = None


        dill.dump(uplift_class_copy, file = open(file_location + '/mr_uplift_class.pkl', "wb"))

        model.save(file_location + '/mr_uplift_model.h5')



    def load(self, file_location):
        """Loads MRUplift Class from location.

        Args:
          file_location (str): File location to load data
        Returns:
          Updated Uplift Class
        """

        uplift_class = dill.load(open(file_location + '/mr_uplift_class.pkl', "rb"))
        uplift_class.model = load_model(file_location + '/mr_uplift_model.h5')
        self.__dict__.update(uplift_class.__dict__)


    def predict_optimal_treatments(self, x, objective_weights=None, treatments=None,
                                   calibrator=False, response_transformer = False,
                                   use_propensity_score_cutoff = True):
        """Calculates optimal treatments of model output given explanatory
            variables and weights

        Args:
          x (np.array): new data to predict. Will use test data if not given
          objective_weights (np.array): set of weights of length num_responses to maximize.
            is required for multi output decisions
          treatments (np.array): Treatments to predict on. If none assigned then
          original training treatments are used.
          calibrator (boolean): If true will use the trained calibrator to transform
          responses. Otherwise will use the response inverse transformer
          response_transformer (boolean): If true will use the trained scaler to transform
            responses. I've noticed that using this in production degrades performance
            becuase model optimizes scaled data.
          use_propensity_score_cutoff(boolean): If false it will not mask treatments for predictions based on propensity
          scores.

        Returns:
          Optimal Treatment Values
        """

        x = np.array(x).astype('float')

        if treatments is None:
            treatments = self.unique_t

        if self.propensity_model is not None and use_propensity_score_cutoff:

            prob_tmts = self.propensity_model.predict_proba(self.x_ss.transform(x))+np.finfo(float).eps
            propensity_scores = pd.DataFrame(1/prob_tmts)

            propensity_scores.columns = self.propensity_model.classes_

            unique_t = reduce_concat(treatments)
            propensity_scores = propensity_scores[unique_t]

            mask_tmt_locations = np.array((propensity_scores < self.propensity_score_cutoff)*1)
        else:
            mask_tmt_locations = None

        ice = self.predict_ice(x, treatments, calibrator, response_transformer = response_transformer)

        if self.num_responses > 1:

            best_treatments = get_best_tmts(objective_weights, ice, treatments,
                mask_tmt_locations = mask_tmt_locations)
        else:

            if self.propensity_model is not None and use_propensity_score_cutoff:
                ice = ice[:,:,0].T
                ice = softmax(ice)*mask_tmt_locations
                ice = ice.T

            else:
                ice = ice[:,:,0]

            best_treatments = treatments[np.argmax(ice, axis=0)]

        return best_treatments

    def calibrate(self, response_transformer = False, treatments=None):
        """(Experimental)
        This fits a calibrator on training dataset. This of the form
        y = b0y_pred_0*t_0+b1*y_pred_1*t_1 + ... + b_num_tmts*y_pred_numtmts*t_num_tmts for all treatments.
        response_transformer (boolean): If true will use the trained scaler to transform
          responses. I've noticed that using this in production degrades performance
          becuase model optimizes scaled data.

        Args:
            None
        Returns:
          None
        """
        x_train, x, y_train, y, t_train, t = train_test_split(
            self.x, self.y, self.t, test_size=self.test_size,
            random_state=self.random_state)

        if t_train.shape[1] > 1:
            t_train = reduce_concat(t_train)
        t_train = t_train.reshape(-1, 1)

        ice = self.predict_ice(x_train, treatments, response_transformer)

        calib = UpliftCalibration()
        calib.fit(ice, y_train, t_train)
        calib.uplift_scores()

        self.calibrator = calib


    def permutation_varimp(self, objective_weights=None, x=None, treatments=None, calibrator=False,
        num_sample = 10000):
        """Variable importance metrics. This is based on permutation tests. For variable this permutes the column
        and then predicts and finds the optimal value given a set of weights. For each user it compares the optimal treatment of
        permuted column data with optimal treatment of non-permuted data and averages the result. The output is an index of how often
        the permuted column disagrees with unpermuted columns.

        Args:
          objective_weights (np.array): set of weights of length num_responses to maximize.
            is required for multi output decisions
          x (np.array): new data to predict. Will use test data if not given
          treatments (np.array): Treatments to predict on. If none assigned then
            original training treatments are used.
          calibrator (boolean): If true will use the trained calibrator to transform
            responses. Otherwise will use the response inverse transformer
          num_sample (int): Number of observations to sample for calculations. Used to reduce
            time of function.
        Returns:
          df of variable importance metrics
        """
        if x is None:
            x_train, x, y_train, y, t_train, t = train_test_split(
                self.x, self.y, self.t, test_size=self.test_size,
                random_state=self.random_state)

        x = np.array(x).astype('float')
        t = np.array(t).astype('float')
        y = np.array(y).astype('float')

        #subset number of observations
        x = x[:num_sample]

        original_decisions = self.predict_optimal_treatments(x,
        objective_weights=objective_weights, treatments=treatments,
                                       calibrator=calibrator)
        varimps = []
        for p in range(x.shape[1]):

            shuffled_index = np.arange(x.shape[0])
            np.random.shuffle(shuffled_index)

            shuffled_index = shuffled_index

            x_copy = x.copy()
            x_copy[:,p] = x_copy[:,p][shuffled_index]

            temp_decisions = self.predict_optimal_treatments(x_copy,
                objective_weights=objective_weights, treatments=treatments,
                calibrator=calibrator)
            temp_varimp = (original_decisions == temp_decisions).mean()

            varimps.append(temp_varimp)

        #make varimps a 'larger number -> more important' metric
        varimps = 1 - np.array(varimps)

        varimps_pd = pd.DataFrame(np.array(varimps))
        varimps_pd.columns = ['permutation_varimp_metric']

        if self.x_names is not None:
            varimps_pd['var_names'] = self.x_names

        return varimps_pd

    def get_random_erupts(self, x = None, y = None, t = None, objective_weights = None,
        treatments = None, calibrator = None, random_seed = 22,
        response_transformer = False, propensity_score_cutoff = 100):
        """OOS metric calculation for full range of a ranom set of objective weights.
        Idea is to calculate full range of objective functions. Here each observation
        is assigned a random objective function and the ERUPT is calculated on this.

        Args:
          x (np.array): new data to predict. Will use test data if not given
          y (np.array): responses
          t (np.array): treatments
          objective_weights (np.array): of dim (num_observations by num_response_variables).
          if none is assigned it randomly create weights
          treatments (np.array): treatments to use in erupt calculations
          calibrator (boolean): If true will use the trained calibrator to transform
            responses. Otherwise will use the response inverse transformer
          random_seed (int): seed for random weights matrix if none are assigned
          response_transformer (boolean): If true will use the trained scaler to transform
            responses. I've noticed that using this in production degrades performance
            becuase model optimizes scaled data.

        Returns:
          mean and standardization of ERUPT
        """

        if treatments is None:
            treatments = self.unique_t

        if x is None:
            x_train, x, y_train, y, t_train, t = train_test_split(
                self.x, self.y, self.t, test_size=self.test_size,
                random_state=self.random_state)

        x = np.array(x).astype('float')
        t = np.array(t).astype('float')
        y = np.array(y).astype('float')

        if objective_weights is None:
            objective_weights = get_random_weights(y, random_seed)

        str_t, str_unique_treatments = treatments_to_text(t, treatments)
        str_t_series = pd.Series(str_t)

        if self.propensity_model is not None:

            prob_tmts = self.propensity_model.predict_proba(self.x_ss.transform(x))+np.finfo(float).eps
            propensity_scores = pd.DataFrame(1/prob_tmts)
            propensity_scores.columns = self.propensity_model.classes_

            mask_tmt_locations = np.array((propensity_scores < propensity_score_cutoff)*1)


            observation_weights = propensity_scores.lookup(str_t_series.index, str_t_series.values)

        else:
            mask_tmt_locations = None
            observation_weights = get_weights(str_t)


        tmt_effects = pd.DataFrame(y)
        tmt_effects['tmt'] = str_t
        tmt_effects_mean = tmt_effects.groupby('tmt').mean()

        utility_by_best_global_tmt = np.array([(objective_weights*x).sum(axis=1) for x in tmt_effects_mean])
        best_single_tmt = tmt_effects_mean.index.values[utility_by_best_global_tmt.argmax(axis = 0)]

        utility = (objective_weights*y).sum(axis=1)

        y_temp = np.concatenate([y, utility.reshape(-1,1)], axis = 1)


        optim_tmt = self.predict_optimal_treatments(x,
                                                    objective_weights=objective_weights,  treatments=treatments,
                                                    calibrator=calibrator, response_transformer = response_transformer)

        new_y = (objective_weights*y).sum(axis = 1).reshape(-1,1)

        erupt_new_y = erupt(new_y, t, optim_tmt, weights = observation_weights)
        erupt_base_tmt = erupt(new_y, str_t_series, best_single_tmt, weights = observation_weights.reshape(-1,1))

        erupt_new_y['type'] = 'model'
        erupt_base_tmt['type'] = 'ate'

        erupts = pd.concat([erupt_new_y, erupt_base_tmt])


        return(erupts)



import numpy as np
import pytest

from mr_uplift.dataset.data_simulation import get_no_noise_data, get_simple_uplift_data
from mr_uplift.calibrate_uplift import UpliftCalibration
from mr_uplift.erupt import weighted_avg_and_std, erupt,\
    get_best_tmts, get_weights


class TestCalibrate(object):

    @pytest.fixture(scope='class')
    def calib(self):
        num_obs = 100
        y_no_noise, x_no_noise, tmt_no_noise = get_no_noise_data(num_obs)
        fake_counters = np.array(
            [np.array([np.zeros(num_obs), np.zeros(num_obs)]).T, x_no_noise])

        calib = UpliftCalibration()
        calib.fit(fake_counters, y_no_noise, tmt_no_noise)

        return calib

    def test_calibration(self, calib):

        assert np.allclose(np.sum(calib.calibrated_intercepts), 0)
        assert np.allclose(calib.calibrated_coefs[:, 1], np.array([2.0, 1.0]))

    def test_calibrate_new_data(self, calib):
        num_obs = 100
        true_coefs = np.array([2.0, 1.0])
        y_no_noise, x_no_noise, tmt_no_noise = get_no_noise_data(num_obs)
        fake_counters = np.array(
            [np.array([np.zeros(num_obs), np.zeros(num_obs)]).T, x_no_noise])

        calib_x = calib.transform(fake_counters)
        assert np.allclose((calib_x[1] / x_no_noise).mean(axis=0), true_coefs)

    def test_uplift_scores(self, calib):
        true_coefs = np.array([[1., 0.], [1., 0.]])
        calib.uplift_scores()
        assert np.allclose(calib.regression_results, true_coefs)



import numpy as np
import pandas as pd
import random

def softmax(x):
    """Compute softmax values for each sets of scores in x."""
    e_x = np.exp(x - np.max(x, axis = 1).reshape(-1,1))
    return e_x / e_x.sum(axis = 1).reshape(-1,1)

def weighted_avg_and_std(values, weights):
    """computes weighted averages and stdevs
    Args:
        values (np.array): variable of interest
        weight (np.array): weight to assign each observation
    Returns:
        list of weighted mean and stdev
    """

    average = np.average(values, weights=weights)

    variance = np.average((values - average)**2, weights=weights)
    ess = np.sum(weights)**2 / np.sum(weights**2)

    return [average, np.sqrt(variance) / np.sqrt(ess)]


def erupt(y, tmt, optim_tmt, weights=None, names=None):
    """Calculates Expected Response Under Proposed Treatments
    Args:
        y (np.array): response variables
        tmt (np.array): treatment randomly assigned to observation
        optim_tmt (np.array): proposed treatment model assigns
        weights (np.array): weights for each observation. useful if treatment
            was not uniformly assigned
        names (list of str): names of response variables
    Returns:
        Expected means and stdevs of proposed treatments
    """

    if weights is None:
        weights = np.ones(y.shape[0])

    equal_locs = np.where(optim_tmt == tmt)[0]

    erupts = [weighted_avg_and_std(y[equal_locs][:,x].reshape(-1,1),
            weights[equal_locs].reshape(-1,1)) for x in range(y.shape[1])]

    erupts = pd.DataFrame(erupts)
    erupts.columns = ['mean', 'std']

    if names is not None:
        erupts['response_var_names'] = names
    else:
        erupts['response_var_names'] = [
            'var_' + str(x) for x in range(y.shape[1])]

    return erupts


def get_best_tmts(objective_weights, ice, unique_tmts, mask_tmt_locations = None):
    """Calulcates Optimal Treatment for a set of counterfactuals and weights.
    Args:
        objective_weights (np.array): list of weights to maximize in objective function
        ice (np.array): 3-d array of counterfactuals.
            num_tmts x num_observations x num_responses
        unique_tmts (np.array): list of treatments
        mask_tmt_locations (np.array): 2-d array of observations x treatments.
            1 if allowed 0 if user shouldn't receive particular treatment.
    Returns:
        Optimal Treatment with specific objective function
    """

    weighted_ice = [x * objective_weights for x in ice]
    sum_weighted_ice = np.array(weighted_ice).sum(axis=2).T

    if mask_tmt_locations is not None:
        sum_weighted_ice = softmax(sum_weighted_ice*mask_tmt_locations - (1-mask_tmt_locations)*10**6)

    max_values = sum_weighted_ice.argmax(axis=1)

    best_tmt = np.array(unique_tmts)[np.array(max_values)]

    return(best_tmt)


def get_weights(tmts):
    """Calculates weights to apply to tmts. Weight is inversely proportional to
        how common each weight is.
    Args:
        tmts (np.array): treatments randomly assigned to
    Returns:
        array of length tmts that is weight to assign in erupt calculation
    """
    tmts_pd = pd.DataFrame(tmts)
    tmts_pd.columns = ['tmt']
    weight_mat = pd.DataFrame(pd.DataFrame(
        tmts).iloc[:, 0].value_counts() / len(tmts))
    weight_mat['tmt'] = weight_mat.index
    weight_mat.columns = ['weight', 'tmt']
    tmts_pd = tmts_pd.merge(weight_mat, on='tmt', how='left')
    observation_weights = 1.0 / np.array(tmts_pd['weight']).reshape(-1, 1)

    return observation_weights


def get_erupts_curves_aupc(y, tmt, ice, unique_tmts, objective_weights,
                        mask_tmt_locations=None, observation_weights=None,
                           names=None):
    """Calculates optimal treatments and returns erupt
    Args:
        y (np.array): response variables
        tmt (np.array): treatment randomly assigned to observation
        ice (np.array): 3-d array of counterfactuals.
            num_tmts x num_observations x num_responses
        objective_weights (np.array): list of weights to maximize in objective
        function
        names (list of str): names of response variables
        mask_tmt_locations (np.array): 2-d array of observations x treatments.
            1 if allowed 0 if user shouldn't receive particular treatment.

    Returns:
        Dataframe of ERUPT metrics for each response variable for a given set
        of objective weights
    """

    tmt_effects = pd.DataFrame(y)
    tmt_effects['tmt'] = tmt
    tmt_effects_mean = tmt_effects.groupby('tmt').mean()

    all_erupts = []
    all_distributions = []

    if observation_weights is None:
        observation_weights = np.ones(y.shape[0])

    if mask_tmt_locations is None:
        mask_tmt_locations = np.ones(y.shape[0]*len(unique_tmts)).reshape(y.shape[0], len(unique_tmts))



    for obj_weight in objective_weights:

        best_single_tmt = tmt_effects_mean.index.values[tmt_effects_mean.multiply(obj_weight, axis = 1).sum(axis=1).argmax()]
        utility = (obj_weight.reshape(1,-1)*y).sum(axis=1)

        y_temp = np.concatenate([y, utility.reshape(-1,1)], axis = 1)

        optim_tmt = get_best_tmts(obj_weight, ice, unique_tmts, mask_tmt_locations)
        random_tmt = optim_tmt.copy()[
            np.random.choice(
                len(optim_tmt),
                len(optim_tmt),
                replace=False)]

        str_obj_weight = ','.join([str(q) for q in obj_weight])

        if names is not None:
            names = np.append(names, 'utility')

        erupts = erupt(y_temp, tmt, optim_tmt, weights=observation_weights,
                       names=names)

        erupts_random = erupt(y_temp, tmt, random_tmt, weights=observation_weights,
                              names=names)

        erupts_best_single_tmt = erupt(y_temp, tmt, np.repeat(best_single_tmt, y.shape[0]), weights=observation_weights,
                              names=names)



        erupts_random['weights'] = str_obj_weight
        erupts['weights'] = str_obj_weight
        erupts_best_single_tmt['weights'] = str_obj_weight


        erupts_random['assignment'] = 'random'
        erupts['assignment'] = 'model'
        erupts_best_single_tmt['assignment'] = 'best_single_tmt'

        erupts['treatment'] = '-1'
        erupts_random['treatment'] = '-1'
        erupts_best_single_tmt['treatment'] = best_single_tmt

        erupts = pd.concat([erupts, erupts_random, erupts_best_single_tmt], axis=0)

        dists = pd.DataFrame(optim_tmt).iloc[:, 0].value_counts()
        dist_treatments = pd.DataFrame(np.array(dists).reshape(-1, 1))
        dist_treatments['tmt'] = dists.index
        dist_treatments['weights'] = str_obj_weight
        dist_treatments['percent_tmt'] = np.array(dists) / np.array(dists).sum()
        dist_treatments.columns = ['num_observations', 'tmt', 'weights',
                                   'percent_tmt']

        all_erupts.append(erupts)
        all_distributions.append(dist_treatments)


    for t in unique_tmts:

        t_rep = np.repeat(t, y.shape[0]).reshape(-1,1)
        str_obj_weight = str(t)

        erupts = erupt(y, tmt.reshape(-1,1), t_rep, weights=observation_weights,
                       names=names)

        erupts['weights'] = '-1'
        erupts['treatment'] = str_obj_weight
        erupts['assignment'] = 'ate'

        all_erupts.append(erupts)

    all_erupts = pd.concat(all_erupts)
    all_distributions = pd.concat(all_distributions)

    return all_erupts, all_distributions



import numpy as np
import pandas as pd
from sklearn.linear_model import LinearRegression


def get_reg_scores(ice, y, tmt):
    """Increase in Scores of using HETE vs control
    Args:
        ice (np array): Individual Conditional Expections 2-d array with
            num_observations, num_tmts.
        This is the counterfactuals for all treatment, response variable pairs.
        y (np array): response variables
        tmt (np array): treatment column
    Returns:
        Builds and returns regression of form y = b0 + b1*y_0_hat +
        b2*(y_1_hat - y_0_hat) + ... + bp*(y_p_hat - y_0_hat)
        and a regression of form y = b0 + b1*y_0_hat.
        It returns in sample scores of each model
    """

    hete = (ice - ice[0, :])

    # if tmt is a binary get_dummies will not expand. turn into string and it
    # will
    tmt = [str(x) for x in tmt]
    temp_data = pd.DataFrame(hete.T * pd.get_dummies(tmt))
    temp_data = temp_data.iloc[:, 1:]
    temp_data['base'] = ice[0, :]

    reg = LinearRegression()
    reg.fit(temp_data, y)

    reg_base = LinearRegression()
    reg_base.fit(temp_data['base'].values.reshape(-1, 1), y)

    score_hete = reg.score(temp_data, y)
    score_base = reg_base.score(
        temp_data['base'].values.reshape(-1, 1), y.reshape(-1, 1))

    return [score_hete, score_base]


def build_calibration_models(ice, y, tmt):
    """Returns Calibrated HETE Models.
    Args:
      ice (np array): Individual Conditional Expections 2-d array with
        num_observations, num_tmts.
      This is the counterfactuals for all treatment, response variable pairs.
      y (np array): response variables
      tmt (np array): treatment column
    Returns:
      Builds and returns regression of form y = b0 + b1*y_0_hat +
      b2*y_1_hat + ... + b(p+1)*(tmt == 0) + ...
    """

    temp_data = pd.DataFrame(ice.T * pd.get_dummies(tmt.reshape(-1,)))
    temp_dummies = pd.get_dummies(tmt.reshape(-1,))
    temp_data = pd.concat([temp_data, temp_dummies], axis=1)

    reg = LinearRegression(fit_intercept=False)
    reg.fit(temp_data, y)

    return reg


class UpliftCalibration:
    """A calibration class for uplift models. Currently has a .fit and
    .transform method to get calibrated effects for each treatment.
    """

    def fit(self, ice, y, tmt):
        """Fits Calibrated HETE Models.
        Args:
          ice (np array): Individual Conditional Expections 2-d array with
            num_observations, num_tmts.
          This is the counterfactuals for all treatment, response variable pairs.
          y (np array): response variables
          tmt (np array): treatment column
        Returns:
          Builds and returns regression of form y = b0 + b1*y_0_hat +
          b2*(y_1_hat - y_0_hat) + ... + bp*(y_p_hat - y_0_hat)
        """

        self.ice = ice
        self.y = y
        self.tmt = tmt
        self.num_tmts = len(np.unique(tmt))
        self.num_responses = y.shape[1]

        self.regs = [build_calibration_models(
            self.ice[:, :, index],
            self.y[:, index],
            self.tmt)
            for index in range(self.num_responses)]

        calibrated_coefs = np.array(
            [self.regs[x].coef_ for x in range(self.num_responses)])
        self.calibrated_coefs = calibrated_coefs[:, :self.num_tmts]
        self.calibrated_intercepts = calibrated_coefs[:, self.num_tmts:]

    def transform(self, new_ice):
        """Transforms counterfactual predictions to be calibrated
        Args:
          new_ice (np array): Individual Conditional Expections 2-d array with
            num_observations, num_tmts.
        Returns:
          Calibrated counterfactuals. Thef first treatment will be considered
          the 'base' and is set to zero. All other counterfactuals are relative
          to this treatment.
        """

        counters_scaled_diff_mats_calibrated = [
            self.calibrated_intercepts.T + new_ice[:, x, :] * self.calibrated_coefs.T for x in range(new_ice.shape[1])]
        counters_scaled_diff_mats_calibrated = np.swapaxes(
            np.array(counters_scaled_diff_mats_calibrated), 1, 0)

        return counters_scaled_diff_mats_calibrated

    def uplift_scores(self):
        """Calculates increase in Scores of using HETE vs control
        Returns:
          Two scores for each response variable. One with all treatments and
          one with only first treatment. This should give an idea of how well
          uplift model does for each treatment
        """

        regression_results = [get_reg_scores(
            self.ice[:, :, index],
            self.y[:, index],
            self.tmt)
            for index in range(self.num_responses)]
        regression_results = pd.DataFrame(regression_results)
        regression_results.columns = [
            'with_uplift_effects',
            'without_uplift_effects']

        self.regression_results = regression_results



import numpy as np
import pytest

from mr_uplift.dataset.data_simulation import get_no_noise_data, get_simple_uplift_data, get_observational_uplift_data_1
from mr_uplift.mr_uplift import MRUplift, get_t_data
from mr_uplift.keras_model_functionality import prepare_data_optimized_loss
import sys
import pandas as pd

class TestMRUplift(object):

    def test_get_t_data(self):

        num_obs_1 = 10
        num_obs_2 = 3

        test_1 = get_t_data(0, num_obs_1)
        test_2 = get_t_data(np.array([0, 1]), num_obs_2)

        test_1_values = np.zeros(num_obs_1).reshape(-1, 1)
        test_2_values = np.concatenate([np.zeros(num_obs_2).reshape(-1, 1),
                                        np.ones(num_obs_2).reshape(-1, 1)], axis=1)

        assert np.mean(test_1 == test_1_values) == 1
        assert np.mean(test_2 == test_2_values) == 1

    def test_model_mean_outputs(self):
        true_ATE = np.array([[0, 0], [1, .5]])
        rmse_tolerance = .05
        num_obs = 10000

        y_no_noise, x_no_noise, tmt_no_noise = get_no_noise_data(num_obs)

        uplift_model = MRUplift()
        uplift_model.fit(x_no_noise, y_no_noise, tmt_no_noise.reshape(-1, 1),
                         n_jobs=1)
        oos_ice = uplift_model.predict_ice(response_transformer = True)

        assert np.sqrt(np.mean((oos_ice.mean(axis=1) -true_ATE)**2)) < rmse_tolerance

    def test_model_pred_oos_shapes(self):
        num_obs = 1000

        y, x, t = get_simple_uplift_data(num_obs)

        t = np.concatenate([t.reshape(-1, 1),
        np.random.binomial(1, .5, num_obs).reshape(-1, 1)], axis=1)

        param_grid = dict(num_nodes=[8], dropout=[.1], activation=[
                'relu'], num_layers=[1], epochs=[1], batch_size=[1000])

        uplift_model = MRUplift()
        uplift_model.fit(x, y, t, param_grid = param_grid, n_jobs=1)

        assert uplift_model.predict_ice().shape == (
            np.unique(t, axis=0).shape[0], num_obs * .7, y.shape[1])

        assert uplift_model.predict_ice(x=x).shape == (np.unique(t,axis=0).shape[0],
            num_obs,
            y.shape[1])

        assert uplift_model.get_erupt_curves()

        assert uplift_model.get_erupt_curves(x = x, y = y, t = t)

    def test_model_pred_oos_shapes_single_col_tmt(self):
        num_obs = 1000

        y, x, t = get_simple_uplift_data(num_obs)

        t = t.reshape(-1, 1)

        param_grid = dict(num_nodes=[8], dropout=[.1], activation=[
                                  'relu'], num_layers=[1], epochs=[1], batch_size=[1000])

        uplift_model = MRUplift()
        uplift_model.fit(x, y, t, param_grid = param_grid, n_jobs=1)

        assert uplift_model.predict_ice().shape == (
            np.unique(t, axis=0).shape[0], num_obs * .7, y.shape[1])

        assert uplift_model.predict_ice(x=x).shape == (np.unique(t,axis=0).shape[0],
            num_obs,
            y.shape[1])

        assert uplift_model.get_erupt_curves()

        assert uplift_model.get_erupt_curves(x = x, y = y, t = t)

    def test_model_pred_oos_shapes_single_col_tmt_propensity(self):
        num_obs = 1000

        y, x, t = get_simple_uplift_data(num_obs)

        t = t.reshape(-1, 1)

        param_grid = dict(num_nodes=[8], dropout=[.1], activation=['relu'], num_layers=[2],
                                epochs=[1], batch_size=[100],
                                  alpha = [.5], copy_several_times = [1])

        uplift_model = MRUplift()
        uplift_model.fit(x, y, t, param_grid = param_grid, n_jobs=1,
            optimized_loss = True, use_propensity = True)

        assert uplift_model.predict_ice().shape == (
            np.unique(t, axis=0).shape[0], num_obs * .7, y.shape[1])

        assert uplift_model.predict_ice(x=x).shape == (np.unique(t,axis=0).shape[0],
            num_obs,
            y.shape[1])

        assert uplift_model.get_erupt_curves()

        assert uplift_model.get_erupt_curves(x = x, y = y, t = t)


    def test_prepare_data_optimized_loss_one_col_tmt(self):
        num_obs = 1000

        y, x, t = get_simple_uplift_data(num_obs)
        t = t.reshape(len(t),1)

        unique_treatments = np.unique(t, axis = 0)

        masks = np.ones(num_obs).reshape(num_obs,1)

        x, utility_weights, missing_utility, missing_y_mat, masks, weights = prepare_data_optimized_loss(x,y,t, masks ,unique_treatments)

        assert(utility_weights.shape == (num_obs, y.shape[1]))
        assert(missing_y_mat.shape == (num_obs, unique_treatments.shape[0], y.shape[1]))
        for q in range(unique_treatments.shape[0]):
            assert( ((missing_utility[:,q]==0) == (missing_y_mat[:,q,0] == -999)).mean() ==1 )

    def test_prepare_data_optimized_loss_two_col_tmt(self):
        num_obs = 1000

        y, x, t = get_simple_uplift_data(num_obs)

        t = np.concatenate([t.reshape(-1, 1),
        np.random.binomial(1, .5, num_obs).reshape(-1, 1)], axis=1)
        unique_treatments = np.unique(t, axis = 0)
        masks = np.ones(num_obs*len(unique_treatments)).reshape(num_obs,len(unique_treatments))

        x, utility_weights, missing_utility, missing_y_mat, masks, weights = prepare_data_optimized_loss(x,y,t,masks, unique_treatments)

        assert(utility_weights.shape == (num_obs, y.shape[1]))
        assert(missing_y_mat.shape == (num_obs, unique_treatments.shape[0], y.shape[1]))
        for q in range(unique_treatments.shape[0]):
            assert( ((missing_utility[:,q]==0) == (missing_y_mat[:,q,0] == -999)).mean() ==1 )

    def test_model_optim_mean_outputs(self):
        true_ATE = np.array([[0, 0], [1, .5]])
        rmse_tolerance = .05
        num_obs = 10000
        param_grid = dict(num_nodes=[8], dropout=[.1], activation=['relu'], num_layers=[2], epochs=[30], batch_size=[100])

        y_no_noise, x_no_noise, tmt_no_noise = get_no_noise_data(num_obs)

        uplift_model = MRUplift()
        uplift_model.fit(x_no_noise, y_no_noise, tmt_no_noise.reshape(-1, 1),
                         n_jobs=1, param_grid = param_grid, optimized_loss = False)
        oos_ice = uplift_model.predict_ice(response_transformer = True)

        assert np.sqrt(np.mean((oos_ice.mean(axis=1) - true_ATE)**2)) < rmse_tolerance

    def test_model_get_random_erupts(self):
        true_ATE = np.array([[0, 0], [1, .5]])
        rmse_tolerance = .05
        num_obs = 10000
        param_grid = dict(num_nodes=[8], dropout=[.1], activation=['relu'], num_layers=[2], epochs=[30], batch_size=[100],
                                  alpha = [.5], copy_several_times = [2])

        y_no_noise, x_no_noise, tmt_no_noise = get_no_noise_data(num_obs)

        uplift_model = MRUplift()
        uplift_model.fit(x_no_noise, y_no_noise, tmt_no_noise.reshape(-1, 1),
                         n_jobs=1, param_grid = param_grid, optimized_loss = True)
        oos_re = uplift_model.get_random_erupts()

        uplift_model_propensity = MRUplift()
        uplift_model_propensity.fit(x_no_noise, y_no_noise, tmt_no_noise.reshape(-1, 1),
                         n_jobs=1, param_grid = param_grid,
                         optimized_loss = True, use_propensity = True)
        oos_re_propensity = uplift_model_propensity.get_random_erupts()


        assert oos_re['mean'].iloc[0] > 0
        assert oos_re_propensity['mean'].iloc[0] > 0

    def test_varimp(self):
        num_obs = 10000
        param_grid = dict(num_nodes=[8], dropout=[.1], activation=['relu'], num_layers=[2], epochs=[30], batch_size=[100])

        y, x, t = get_simple_uplift_data(num_obs)

        uplift_model = MRUplift()
        uplift_model.fit(x, y, t.reshape(-1, 1),
                         n_jobs=1, param_grid = param_grid)
        varimp = uplift_model.permutation_varimp(objective_weights = np.array([.7,-.3,0]).reshape(1,-1))

        param_grid = dict(num_nodes=[8], dropout=[.1], activation=['relu'], num_layers=[2], epochs=[30], batch_size=[100],
                                  alpha = [.5], copy_several_times = [2])

        uplift_model_propensity = MRUplift()
        uplift_model_propensity.fit(x, y, t.reshape(-1, 1),
                         n_jobs=1, param_grid = param_grid,
                         optimized_loss = True, use_propensity = True)
        varimp_propensity = uplift_model_propensity.permutation_varimp(objective_weights = np.array([.7,-.3,0]).reshape(1,-1))

        assert varimp['permutation_varimp_metric'].iloc[0]>varimp['permutation_varimp_metric'].iloc[1]
        assert varimp_propensity['permutation_varimp_metric'].iloc[0]>varimp_propensity['permutation_varimp_metric'].iloc[1]


    def test_model_propensity(self):
        num_obs = 10000
        TOLERANCE = .98
        y, x, t, rule_assignment = get_observational_uplift_data_1(num_obs)

        param_grid = dict(num_nodes=[8], dropout=[.1], activation=['relu'], num_layers=[1],
                                epochs=[20], batch_size=[512],
                                  alpha = [.9999,.99], copy_several_times = [1])

        uplift_model = MRUplift()
        uplift_model.fit(x, y[:,0].reshape(-1,1), t, param_grid = param_grid, n_jobs=1,
            optimized_loss = True, use_propensity = True, test_size = 0)
        uplift_model.best_params_net
        y_test, x_test, t_test, rule_assignment_test = get_observational_uplift_data_1(num_obs)


        experiment_groups = np.zeros(num_obs)+2
        experiment_groups[np.where(x_test[:,-2]<.5)[0]] = 1
        experiment_groups[np.where(x_test[:,-2]<.33)[0]] = 0
        experiment_groups[np.where(x_test[:,-1]>.8)[0]] = 3


        optim_treatments_no_cuttoff = uplift_model.predict_optimal_treatments(x = x_test, use_propensity_score_cutoff = False)
        optim_treatments_cuttoff = uplift_model.predict_optimal_treatments(x = x_test, use_propensity_score_cutoff = True)

        optim_treatments_cuttoff_cat = optim_treatments_cuttoff.argmax(axis = 1)
        optim_treatments_no_cuttoff_cat = optim_treatments_no_cuttoff.argmax(axis = 1)

        correct_tmts_1 = np.array([x in [0,1] for x in optim_treatments_cuttoff_cat[np.where(experiment_groups == 0)[0]] ]).mean()
        correct_tmts_2 = np.array([x in [1,2] for x in optim_treatments_cuttoff_cat[np.where(experiment_groups == 1)[0]] ]).mean()
        correct_tmts_3 = np.array([x in [0,2] for x in optim_treatments_cuttoff_cat[np.where(experiment_groups == 2)[0]] ]).mean()
        correct_tmts_4 = np.array([x in [0] for x in optim_treatments_cuttoff_cat[np.where(experiment_groups == 3)[0]] ]).mean()

        correct_tmts_experiment_groups_1 = ((optim_treatments_cuttoff_cat[np.where(experiment_groups == 1)[0]] == 1)  == x_test[np.where(experiment_groups == 1)[0],0]).mean()
        correct_tmts_no_cutoff =  np.mean((optim_treatments_no_cuttoff_cat==1    ) == x_test[:,0])

        assert correct_tmts_1>TOLERANCE
        assert correct_tmts_2>TOLERANCE
        assert correct_tmts_3>TOLERANCE
        assert correct_tmts_4>TOLERANCE

        assert correct_tmts_experiment_groups_1>TOLERANCE

        assert np.array_equal(optim_treatments_cuttoff_cat,optim_treatments_no_cuttoff_cat) is False
        assert correct_tmts_no_cutoff>TOLERANCE



"""
Run setup
"""

from setuptools import setup, find_packages

def parse_description(description):
    """
    Strip figures and alt text from description
    """
    return "\n".join(
        [
        a for a in description.split("\n")
        if ("figure::" not in a) and (":alt:" not in a)
        ])

DISTNAME = "mr_uplift"
VERSION = "0.0.14"
DESCRIPTION = "Machine learning tools for uplift models"
with open("README.rst") as f:
    LONG_DESCRIPTION = parse_description(f.read())
CLASSIFIERS = [
    "Intended Audience :: Developers",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3.5",
    "Programming Language :: Python :: 3.6",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3.8",
    "Topic :: Scientific/Engineering"
    ]
AUTHOR = "Ibotta Inc."
AUTHOR_EMAIL = "machine_learning@ibotta.com"
LICENSE = "Apache 2.0"
DOWNLOAD_URL = "https://github.com/Ibotta/mr_uplift"
PROJECT_URLS = {
    "Source Code": "https://github.com/Ibotta/mr_uplift"
    }
MIN_PYTHON_VERSION = "3.5"

setup(name=DISTNAME,
      version=VERSION,
      description=DESCRIPTION,
      long_description=LONG_DESCRIPTION,
      classifiers=CLASSIFIERS,
      author=AUTHOR,
      author_email=AUTHOR_EMAIL,
      license=LICENSE,
      download_url=DOWNLOAD_URL,
      project_urls=PROJECT_URLS,
      packages=find_packages(),
      install_requires=[
          'numpy>1.17.2',
          'pandas>=0.20.0',
          'tensorflow==2.4.1',
          'keras',
          'dill',
          'sklearn'
      ],
      python_requires=">={0}".format(MIN_PYTHON_VERSION)
      )



import numpy as np
import pandas as pd

def get_simple_uplift_data(num_obs):
    """Creates sample uplift dataset with 3 variables.
    First two variables are of form y_i = x_i*t + e for two responses
    Thrid response is just noise

    Args:
        num_obs (int): number of observations to simulate from
    Returns:
        responses, explanatory variables, and treatment
    """

    tmt = np.random.binomial(1, .5, num_obs)
    x = np.concatenate([np.random.uniform(0, 1, num_obs).reshape(-1, 1),
                        np.random.uniform(0, 1, num_obs).reshape(-1, 1)], axis=1)

    y_1 = tmt * x[:, 0] + np.random.normal(0, .1, num_obs)
    y_2 = tmt * x[:, 1] + np.random.normal(0, .1, num_obs)
    y_3 = np.random.normal(0, 1, num_obs).reshape(-1, 1)

    y = np.concatenate(
        [y_1.reshape(-1, 1), y_2.reshape(-1, 1), y_3.reshape(-1, 1)], axis=1)

    return y, x, tmt


def get_sin_uplift_data(num_obs):
    """Creates sample uplift dataset with 3 variables.
    First two variables are of form y_i = x_i*t + e for two responses
    Thrid response is just noise

    Args:
        num_obs (int): number of observations to simulate from
    Returns:
        responses, explanatory variables, and treatment
    """

    tmt = np.random.binomial(1, .5, num_obs)
    x = np.concatenate([np.random.uniform(0, 1, num_obs).reshape(-1,1),
    np.random.uniform(0, 1, num_obs).reshape(-1,1)], axis = 1)

    y_1 = np.sin(tmt*20*x[:,0]) + np.random.normal(0, .1, num_obs)
    y_2 = np.sin(tmt*20*x[:,1]) + np.random.normal(0, .1, num_obs)
    y_3 =  np.random.normal(0, 1, num_obs).reshape(-1,1)

    y = np.concatenate([y_1.reshape(-1,1), y_2.reshape(-1,1),y_3.reshape(-1,1)], axis = 1)

    return y, x, tmt



def get_no_noise_data(num_obs):
    """Creates sample uplift dataset with 3 variables.
    First two variables are of form y_i = x_i*t for two responses
    Thrid response is just noise

    Args:
        num_obs (int): number of observations to simulate from
    Returns:
        responses, explanatory variables, and treatment
    """

    tmt = np.random.binomial(1, .5, num_obs)
    x = np.concatenate([np.random.uniform(0, 1, num_obs).reshape(-1, 1),
                        np.random.uniform(0, 1, num_obs).reshape(-1, 1)], axis=1)

    noise = np.random.normal(0, .1, num_obs)

    y_1 = 2 * tmt * x[:, 0]
    y_2 = tmt * x[:, 1]

    y = np.concatenate([y_1.reshape(-1, 1), y_2.reshape(-1, 1), ], axis=1)

    return y, x, tmt


def get_no_noise_data_2(num_obs):
    """Creates sample uplift dataset with 3 variables.
    First two variables are of form y_i = x_i*t for two responses
    Thrid response is just noise

    Args:
        num_obs (int): number of observations to simulate from
    Returns:
        responses, explanatory variables, and treatment
    """

    tmt = np.random.choice([0, 1, 2, 3], num_obs,
                           [.25, .25, .25, .25]).reshape(-1, 1)
    x = np.concatenate([
        np.random.uniform(0, 1, num_obs).reshape(-1, 1),
        np.random.uniform(0, 1, num_obs).reshape(-1, 1),
        np.random.uniform(0, 1, num_obs).reshape(-1, 1),
        np.random.uniform(0, 1, num_obs).reshape(-1, 1),

    ], axis=1)

    noise = np.random.normal(0, .1, num_obs)

    y = 1 * (tmt == 0) * x[:, 0].reshape(-1, 1) + (tmt == 0) * 1 + \
        2 * (tmt == 1) * x[:, 1].reshape(-1, 1) + (tmt == 1) * 2 + \
        3 * (tmt == 2) * x[:, 2].reshape(-1, 1) + (tmt == 2) * 3 + \
        4 * (tmt == 3) * x[:, 3].reshape(-1, 1) + (tmt == 3) * 4

    return y, x.T, tmt



def get_observational_uplift_data_1(num_obs, noise = False):
    """Creates sample uplift dataset with an experiment between different 'models'

    Args:
        num_obs (int): number of observations to simulate from
    Returns:
        responses, explanatory variables, and treatment
    """

    segment = np.random.choice(['a','b'], size = num_obs)
    segment = pd.get_dummies(segment)
    x = np.concatenate([np.random.uniform(0, 1, num_obs).reshape(-1, 1),
                        np.random.uniform(0, 1, num_obs).reshape(-1, 1)],
                        axis=1)

    rule_assignment = np.random.choice(['control','assignment_1','assignment_2'], size = num_obs, p=[.1, 0.45, 0.45] )
    tmt = np.zeros(num_obs)
    #model_1
    tmt[np.where((rule_assignment=='assignment_1')*(x[:,0]<.33))[0]] = 1
    tmt[np.where((rule_assignment=='assignment_1')*(x[:,0]>.33))[0]] = 2
    #model_2
    tmt[np.where((rule_assignment=='assignment_2')*(x[:,0]<.5))[0]] = 1
    tmt[np.where((rule_assignment=='assignment_2')*(x[:,0]>.5))[0]] = 2
    #control
    tmt[np.where(x[:,1]>.8)] = 0

    tmt = pd.get_dummies(tmt)

    y = np.array((segment['a'] == 1) * (tmt[1.0])).astype('float')
    y = y + np.array((segment['b'] == 1) * (tmt[2.0])).astype('float')

    y = y - np.array((segment['b'] == 1) * (tmt[1.0])).astype('float')
    y = y - np.array((segment['a'] == 1) * (tmt[2.0])).astype('float')

    y = np.array(y).reshape(-1,1)

    if noise:
        y = y + np.random.normal(0, .1, num_obs)

    x = np.concatenate([np.array(segment), x], axis = 1)

    tmt = np.array(tmt)

    return y, x, tmt, rule_assignment


def get_observational_uplift_data_2(num_obs):
    """Creates sample uplift dataset with 3 variables.
    First two variables are of form y_i = x_i*t + e for two responses
    Thrid response is just noise. Treatment is a function of x_1

    Args:
        num_obs (int): number of observations to simulate from
    Returns:
        responses, explanatory variables, and treatment
    """

    x = np.concatenate([np.random.uniform(0, 1, num_obs).reshape(-1, 1),
                        np.random.uniform(0, 1, num_obs).reshape(-1, 1)], axis=1) - .5

    tmt = np.random.binomial(1, .2+.6*(x[:,0]>0), num_obs)

    y_1 = tmt * x[:, 0] + np.random.normal(0, .1, num_obs)
    y_2 = tmt * x[:, 1] + np.random.normal(0, .1, num_obs)
    y_3 = np.random.normal(0, 1, num_obs).reshape(-1, 1)

    y = np.concatenate(
        [y_1.reshape(-1, 1), y_2.reshape(-1, 1), y_3.reshape(-1, 1)], axis=1)

    return y, x, tmt
